# include <string.h>
# include <SPI.h>

# include "avr/pgmspace.h"

# include "Hands.h"

extern char p_buf[128];
extern char ibuf[32];

/**********************************************************************/
/*   A pointer to the only instance of hands there should be!         */
/**********************************************************************/
Hands* Hands::m_hands = 0;

/**********************************************************************/
/*   Channel  lookup  for  each digit. Array starts with hour 12 and  */
/*   proceeds from 1 through 12                                       */
/**********************************************************************/
const uint8_t  Hands::hour2channel[] = 
{
// 0  1  2   3  4  5  6  7  8   9  10 11 12
   1, 2, 3, 0, 11, 7, 5, 6, 4, 9, 10, 8, 1
//   2, 3, 0, 11, 8, 5, 6, 4, 9, 10, 7, 1, 2
//   2, 3,  4, 1, 12, 9, 6, 7, 5, 10, 11, 8, 2
};   

static bool no_dbg = false;

Hands *Hands::instance()
{
   if ( m_hands == 0 ) 
   {
      m_hands = new Hands;
   }
   return(m_hands);
}

Hands::Hands()
{
   pinMode(DAC_SELECT, OUTPUT);
   digitalWrite(DAC_SELECT, HIGH);
   SPI.begin();
   SPI.beginTransaction(SPISettings(8000000, MSBFIRST, SPI_MODE3));   

   for (uint8_t x = 1; x < 13; x++ ) 
   {
      no_dbg = true;
      setHand(x, 0);
      no_dbg = false;
   }
   m_hours = 0;
   m_minutes = 0;
   m_seconds = 1;
   memset(m_values, 0x00, sizeof(m_values));
   memset(m_values, 0x00, sizeof(m_targets));
}

void Hands::setHand(uint8_t hand, uint8_t value, bool fluid)
{
//   if ( hand > 0 ) 
//   {
//      hand--;
//   }
   if ( fluid ) 
   {
//      Serial << strcpy_P(p_buf, PSTR("setHand: setting target for: ")) << (int)hand;
//      Serial << strcpy_P(p_buf, PSTR(" to value: ")) << (int)value << endl;
      m_targets[hand] = value;
   }
   else
   {
//      Serial << strcpy_P(p_buf, PSTR("setHand: setting target and writing value for: ")) << (int)hand;
//      Serial << strcpy_P(p_buf, PSTR(" to value: ")) << (int)value << endl;
      m_targets[hand] = value;
      m_values[hand] = value;
      writeHand(hand, value);
   }
}

void Hands::forceHand(uint8_t hand, uint8_t value)
{
   if ( hand > 0 ) 
   {
      hand--;
   }
   m_values[hand] = value;
}

void Hands::Refresh()
{
   for (uint8_t hand = 0; hand < 12; hand++) 
   {
      m_values[hand] = 0;
   }
}
      
void Hands::AdjustHands()
{
   uint8_t hand;

   for (hand = 0; hand < 12; hand++) 
   {
      if ( m_values[hand] != m_targets[hand] ) 
      {
         if ( m_values[hand] + ADJUST_HAND_INCREMENT < m_targets[hand] ) 
         {
            m_values[hand] += ADJUST_HAND_INCREMENT;
         }
         else if ( m_values[hand] - ADJUST_HAND_INCREMENT > m_targets[hand] ) 
         {
            m_values[hand] -= ADJUST_HAND_INCREMENT;
         }
         else 
         {
            m_values[hand] = m_targets[hand];
         }
         writeHand(hand, m_values[hand]);
      }
   }
}

void Hands::writeHand(uint8_t hand, uint8_t value)
{
   uint8_t data1;
   uint8_t data2;
   char p_buf[128];

//   Serial << strcpy_P(p_buf, PSTR(" Hands::setHand(): hand: ")) << (int)hand;
   int ival = value;
//   Serial << strcpy_P(p_buf, PSTR(", ival (")) << ival;
   ival = m_settings->handAdjust[hand] * ival / 100;
//   Serial << strcpy_P(p_buf, PSTR(") * m_adjust (")) << (int)m_adjust[hand];
//   Serial << strcpy_P(p_buf, PSTR(") = ")) << ival;
   uint8_t hval = ival;
   uint8_t hand_adj = hour2channel[hand];
   if ( false && ! no_dbg) 
   {
      Serial << strcpy_P(p_buf, PSTR(" Hands::setHand(): hand: ")) << (int)hand;
      Serial << strcpy_P(p_buf, PSTR(", hand: ")) << (int)hand;
      Serial << strcpy_P(p_buf, PSTR(", hand_adj: ")) << (int)hand_adj;
      Serial << strcpy_P(p_buf, PSTR(", value: ")) << (int)value;
      Serial << strcpy_P(p_buf, PSTR(", hval: ")) << (int)hval << endl;
   }
   data1 = hand_adj;   // << 4 | ((value >> 4) & 0x0f);
   data2 = hval;    // & 0x0f;
   digitalWrite(DAC_SELECT, LOW);
   SPI.transfer(data1);
   SPI.transfer(data2);
   digitalWrite(DAC_SELECT, HIGH);
   //   SPI.endTransaction();   
}

void Hands::setHourHand(uint8_t hand, bool fluid)
{
   if ( m_hours != 0 && m_hours != m_minutes ) 
   {
      setHand(m_hours, OFF_LEN, true);
   }
   if ( hand != m_minutes ) 
   {
      setHand(hand, HOUR_LEN, fluid);
   }
   m_hours = hand;
}

void Hands::setMinuteHand(uint8_t hand, bool fluid)
{
   char p_buf[128];
   if ( false ) 
   {
      Serial << strcpy_P(p_buf, PSTR("Hands::setMinuteHand(): hand: ")) << (int)hand;
      Serial << strcpy_P(p_buf, PSTR(", m_minutes: ")) << (int)m_minutes;
      Serial << strcpy_P(p_buf, PSTR(", m_hours: ")) << (int)m_hours << endl;
   }
   if ( m_minutes != 0 && m_minutes != m_hours ) 
   {
      setHand(m_minutes, OFF_LEN, true);
   }
   setHand(hand, MINUTE_LEN, fluid);
   m_minutes = hand;
}

void Hands::setSecondHand(uint8_t hand, uint8_t qty, bool fluid)
{
   uint8_t skip;
   if ( qty == 2 || qty == 3 || qty == 4 || qty == 6 ) 
   {
      skip = 12 / qty;
   }
   else if (qty == 5)
   {
      skip = 2;
      qty = 6;
   }
   else 
   {
      qty = 1;
   }
   for (uint8_t x = 0; x < 12; x++) 
   {
      if ( x != m_minutes && x != m_hours ) 
      {
         setHand(x, OFF_LEN, true);
      }
   }

   m_seconds = hand;
   for (uint8_t x = qty; x > 0; x--)
   {
//      if ( hand == 0 ) 
//      {
//         setHand(12, SECOND_LEN, fluid);
//      }
//      else 
//      {
         setHand(hand, SECOND_LEN, fluid);
//      }
      hand = ((hand + skip) % 12);
   }
}

