#ifndef HANDS_H
#define HANDS_H

# include "core/HardwareSerial.h"
# include "esquare.h"

class Hands 
{
public:
   enum 
   {
      HOUR_LEN                = 60,
      MINUTE_LEN              = 100,
      SECOND_LEN              = 34,
      OFF_LEN                 = 23
   };

   Hands();
   static Hands *instance();

   void AdjustHands();
   void forceHand(uint8_t hand, uint8_t value);
   void Refresh();
   void setHand(uint8_t hand, uint8_t value, bool fluid = true);
   void setHourHand(uint8_t hand, bool fluid = true);
   void setMinuteHand(uint8_t hand, bool fluid = true);
   void setSecondHand(uint8_t hand, uint8_t qty, bool fluid = true);
   void writeHand(uint8_t hand, uint8_t value);

   void setESquare(ESquareSpace_t *settings)
   {
      m_settings = settings;
   }

   uint8_t seconds()
   {
      return(m_seconds);
   }

   uint8_t minutes()
   {
      return(m_minutes);
   }

   uint8_t hours()
   {
      return(m_hours);
   }

   void incrementHourHand()
   {
      setHourHand(m_hours + 1);
   }

   void incrementMinuteHand()
   {
      setMinuteHand(m_minutes + 1);
   }

   void handAdjust(uint8_t hand, int8_t incr)
   {
      char p_buf[128];
      m_settings->handAdjust[hand] += incr;
      Serial << strcpy_P(p_buf, PSTR(" Hands::handAdjust(): hand: ")) << (int)hand;
      Serial << strcpy_P(p_buf, PSTR(", incr: ")) << (int)incr;
      Serial << strcpy_P(p_buf, PSTR(", handAdjust = ")) << (int)m_settings->handAdjust[hand];
   }
protected:
private:
   enum 
   {
      DAC_SELECT              = 10,
      ADJUST_HAND_INCREMENT   = 2,
   };

   static const uint8_t hour2channel[13];
   uint8_t              m_minutes;
   uint8_t              m_seconds;
   uint8_t              m_hours;
   uint8_t              m_values[13];
   uint8_t              m_targets[13];
   uint8_t              m_adjust[13]; 
   ESquareSpace_t*      m_settings;
   static Hands*        m_hands;
};

#endif
