#define  DISPLAY_OFF

# include <stdint.h>
# include <avr/interrupt.h>

# include "datetime.h"
# include "IOPort.h"

#ifndef sbi
#define sbi(sfr, bit) (_SFR_BYTE(sfr) |= _BV(bit))
#endif

OutputPort dbg1;
OutputPort dbg2;
OutputPort dbg3;
OutputPort dbg4;
OutputPort gen_60hz;
InputPort  irq_60hz;

extern uint32_t systemSeconds;
extern volatile unsigned long timer0_millis;
extern char p_buf[128];
extern char ibuf[32];
extern bool dbg;

Hands*  Clock::m_hands = NULL;
long f_cpu = F_CPU;

static uint16_t array[7];
static uint16_t hand_count = 0;
static volatile uint32_t zcross = 0;

uint32_t Clock::m_lineFreq    = 32768;
uint32_t Clock::m_lineFreq_16 = 32768 / 16;
uint16_t Clock::m_discCount   = 0;

//SIGNAL(INT0_vect)
void countTime()
{
   zcross++;
   hand_count++;

   if ( Clock::m_hands != NULL && hand_count == 900 ) 
   {
      Clock::m_hands->AdjustHands();
      hand_count = 0;
   }
   if ( (zcross % (32768 / 32)) == 0 ) 
   {
      switches.update();
   }
   if ( zcross == Clock::m_lineFreq ) 
   {
//      gen_60hz.on();
      zcross = 0;
      currentDate.increment();
//      gen_60hz.off();
   }
}

Clock::Clock() :
       m_dcpower(false),
       m_initialized(false),
       m_updateNeeded(false),
       m_discover(true)
{
   /***********************************************/
   /*   Enable INT0 (the line freq in) interrupt  */
   /*   to fire on both edges.                    */
   /***********************************************/
   EICRA &= ~(_BV(ISC01));
   EICRA |= _BV(ISC00);
   EIMSK |= _BV(INT0);


   gen_60hz.SetPort(IOPort::PORT_D, _BV(4));
   dbg1.SetPort(IOPort::PORT_D, _BV(6));
   dbg2.SetPort(IOPort::PORT_D, _BV(5));
   dbg3.SetPort(IOPort::PORT_D, _BV(4));
   dbg4.SetPort(IOPort::PORT_C, _BV(3));

   dbg1.off();
   dbg2.off();
   dbg3.off();
   dbg4.off();

//   attachInterrupt(0, countTime, CHANGE); // Attach the routine for counting zero crossings
   attachInterrupt(0, countTime, RISING); // Attach the routine for counting zero crossings
   irq_60hz.SetPort(IOPort::PORT_D, _BV(2));
   irq_60hz.on();
}

void Clock::init()
{
   if (dbg) Serial << __PRETTY_FUNCTION__ << "Setting clock to rtc time" << endl;
   setTime(readClock());
   m_rtc.EnableTrickleCharge(DS1339::TC_250);
   m_initialized = true;
}

void Clock::delay(uint16_t msecs)
{
   uint16_t lf2ms = m_lineFreq / 1000;
   int32_t expire_time = (zcross + (msecs * lf2ms)) % m_lineFreq;
   if (dbg) Serial << __PRETTY_FUNCTION__ << "expire_time:" << expire_time << endl;
   dbg1.on();
   while (zcross != expire_time);
   dbg1.off();
}

void Clock::updateFromRTC()
{
   if (dbg) Serial << __PRETTY_FUNCTION__ << "Setting clock to rtc time" << endl;
   setTime(readClock());
}


void Clock::readClock(uint16_t *dt)
{
   uint8_t sec, min, hour, day, dow, month;
   uint16_t year;

   m_rtc.readTime(&sec, &min, &hour, &day, &dow, &month, &year);
   dt[Second] = sec;
   dt[Minute] = min;
   dt[Hour] = hour;
   dt[Day] = day;
   dt[DayOfWeek] = dow;
   dt[Month] = month;
   dt[Year] = year;
}

time_t Clock::readClock()
{
   uint8_t sec, min, hour, day, dow, month;
   uint16_t year;

   m_rtc.readTime(&sec, &min, &hour, &day, &dow, &month, &year);
//   Serial << __PRETTY_FUNCTION__ << " sec: " << (uint16_t)sec; 
//   Serial << ", min: " << (uint16_t)min; 
//   Serial << ", hour: " << (uint16_t)hour; 
//   Serial << ", day: " << (uint16_t)day; 
//   Serial << ", month: " << (uint16_t)month; 
//   Serial << ", year: " << (uint16_t)year; 
//   Serial << ", dow: " << (uint16_t)dow; 
   month--;
   time_t my_time = makeTime(sec, min, hour, day, month, year);
//   Serial << ", time_t : " << (time_t)my_time << endl; 
   return(my_time);
}

void Clock::writeClock(uint16_t *dt)
{
   uint16_t month = dt[Month] - 1;
   time_t time = makeTime(dt[Second], dt[Minute], dt[Hour], dt[Day], month, dt[Year]);
   if (dbg) Serial << __PRETTY_FUNCTION__ << " sec: " << (uint16_t)dt[Second] << endl; 
   if (dbg) Serial << __PRETTY_FUNCTION__ << " min: " << (uint16_t)dt[Minute] << endl; 
   if (dbg) Serial << __PRETTY_FUNCTION__ << " hour: " << (uint16_t)dt[Hour] << endl; 
   if (dbg) Serial << __PRETTY_FUNCTION__ << " day: " << (uint16_t)dt[Day] << endl; 
   if (dbg) Serial << __PRETTY_FUNCTION__ << " month: " << (uint16_t)month << endl; 
   if (dbg) Serial << __PRETTY_FUNCTION__ << " year: " << (uint16_t)dt[Year] << endl; 
   if (dbg) Serial << __PRETTY_FUNCTION__ << " time: " << (uint16_t)time << endl; 
   writeClock(time);
//   Serial << __PRETTY_FUNCTION__ << " time: " << (uint16_t)time << endl; 
//   Serial << __PRETTY_FUNCTION__ << "m_seconds = " << m_seconds << endl;
}

void Clock::writeClock(time_t time)
{
   uint8_t sec, min, hour, day, dow, month, year;
   uint16_t iyear;

//   Serial << __PRETTY_FUNCTION__ << "m_seconds = " << m_seconds << endl;
   localTime(&time, &sec, &min, &hour, &day, &dow, &month, &year);
   month++;
   year += 1900;
   m_rtc.setTime(sec, min, hour, day, dow, month, year);
}

time_t Clock::getTime()
{
   if ( ! m_initialized ) 
      init();
//   Serial << __PRETTY_FUNCTION__ << "m_seconds = " << m_seconds;
   return(m_seconds);
}

void Clock::getTime(uint16_t *dt)
{
   uint8_t sec, min, hour, day, dow, month, year;
   uint16_t iyear;

   if ( ! m_initialized ) 
      init();
   time_t my_seconds = m_seconds;

   localTime(&my_seconds, &sec, 
                       &min, 
                       &hour, 
                       &day, 
                       &dow, 
                       &month, 
                       &year);

   iyear = year + 1900;
//   Serial << __PRETTY_FUNCTION__ << strcpy_P(p_buf, PSTR("my_seconds = ")) << my_seconds;
//   Serial << strcpy_P(p_buf, PSTR(", sec = ")) << (int)sec;
//   Serial << strcpy_P(p_buf, PSTR(", min = ")) << (int)min;
//   Serial << strcpy_P(p_buf, PSTR(", hour = ")) << (int)hour;
//   Serial << strcpy_P(p_buf, PSTR(", day = ")) << (int)day;
//   Serial << strcpy_P(p_buf, PSTR(", dow = ")) << (int)dow;
//   Serial << strcpy_P(p_buf, PSTR(", month = ")) << (int)month;
//   Serial << strcpy_P(p_buf, PSTR(", year = ")) << (int)year;
//   Serial << endl;

   dt[Second]     = sec;
   dt[Minute]     = min;
   dt[Hour]       = hour;
   dt[Day]        = day;
   dt[DayOfWeek]  = dow;
   dt[Month]      = month;
   dt[Year]       = iyear;
}

void Clock::setTime(uint16_t *dt)
{
   zcross = 0;
   m_seconds = makeTime(dt[Second], dt[Minute], dt[Hour], dt[Day], dt[Month], dt[Year]);
}

void Clock::setTime(time_t time)
{
   zcross = 0;
   m_seconds = time;
}

bool Clock::updateClock()
{
   bool rv = m_updateNeeded;
   if ( m_updateNeeded ) 
   {
      m_updateNeeded = false;
      writeClock(m_seconds);
   }
   return(rv);
}

void Clock::increment()
{
   m_seconds++;
   for (int x = 0; x < INTERVAL_COUNTERS; x++) 
   {
      m_intervalCounter[x]++;
   }
}


