#ifndef CLOCK_H
#define CLOCK_H

# include "Arduino.h"

#include "ds1339.h"
#include "Hands.h"
#include "HardwareSerial.h"
#include "switch.h"

typedef unsigned long time_t;

void countTime();

class Clock 
{
public:
   enum
   {
      ONE_SECOND_TICK         = 0,
      CLOCK_UPDATE_TIMER,

      // Must stay last!!
      INTERVAL_COUNTERS
   };

   enum 
   {
      Second = 0, Minute, Hour, Day, DayOfWeek, Month, Year, NUM_ELEMENTS
   };

   Clock();

   void init();
   void updateFromRTC();

   bool discoveryMode()
   {
      return(m_discover);
   }

   void setDiscoveryMode(bool d)
   {
      m_discover = d;
   }

   uint32_t lineFreq()
   {
      return(m_lineFreq);
   }

   static void delay(uint16_t msecs);

   void readClock(uint16_t *dt);
   time_t readClock();

   void writeClock(uint16_t *dt);
   void writeClock(time_t time);

   void getTime(uint16_t *dt);
   time_t getTime();

   uint16_t getMinutesOfDay()
   {
      uint32_t day_minutes = (uint32_t)(24 * 60) * 60;
      return((m_seconds % (day_minutes)) / 60);
   };

   time_t getInterval(uint8_t counter)
   {
      counter %= INTERVAL_COUNTERS;
//      if ( counter == LOG_TEMPERATURE_INTERVAL ) 
//      {
//         Serial << __PRETTY_FUNCTION__ << "m_intervalCounter[" << (uint32_t)counter << "]: " << m_intervalCounter[counter] << endl;
//      }
      return(m_intervalCounter[counter]);
   };

   void resetInterval(uint8_t counter)
   {
      counter %= INTERVAL_COUNTERS;
//      if ( counter == LOG_TEMPERATURE_INTERVAL ) 
//      {
//         Serial << __PRETTY_FUNCTION__ << "- resetting counter: " << (uint32_t)counter << endl;
//      }
      m_intervalCounter[counter] = 0;
   };

   void setHands()
   {
      m_hands = Hands::instance();
   }
   void setTime(uint16_t *dt);
   void setTime(time_t time);

   bool updateClock();
   void increment();

   virtual void localTime(time_t  *timep,
                          uint8_t *psec,
                          uint8_t *pmin,
                          uint8_t *phour,
                          uint8_t *pday,
                          uint8_t *pwday,
                          uint8_t *pmonth,
                          uint8_t *pyear)
   {
   };

   virtual time_t makeTime(uint8_t sec, 
                           uint8_t min, 
                           uint8_t hour, 
                           uint8_t day, 
                           uint8_t month, 
                           uint16_t year )
   {
      return(0);
   };

   static Hands*     m_hands;
   static uint16_t   m_discCount;
   static uint32_t   m_lineFreq;
   static uint32_t   m_lineFreq_16;

protected:
   bool              m_dcpower;
   unsigned long     m_discoveryStart;
   volatile time_t   m_seconds;
   volatile time_t   m_intervalCounter[INTERVAL_COUNTERS];
   bool              m_initialized;
   bool              m_updateNeeded;
   volatile bool     m_discover;

private:
//   DS1307   m_rtc;
   DS1339   m_rtc;
};

#endif
